@@ -16,7 +16,9 @@ import numpy as np
 
 
 # regular expression of an integer number
-NUMBER_RE = re.compile("^[0-9]*[1-9][0-9]*$")
+# NUMBER_RE = re.compile("^[0-9]*[1-9][0-9]*$")
+NUMBER_RE = re.compile("^[0-9]+$")
+
 LINK_TYPES = ["SEMROLE", "SRLINK"]
 
 LINK_TYPES = ["SEMROLE","SRLINK"]
@@ -52,18 +54,25 @@ class ReadBrat(read.Read):
       Reader to brat file annotations and their text files
     """
 
-    def __init__(self):
+    def __init__(self, lang="en"):
         """
         Load spacy model to read process brat files. Also
         store the files processed.
         """
-        if not(spacy.util.is_package('pt_core_news_lg')):
-            spacy.cli.download('pt_core_news_lg')
-
-        self.nlp = spacy.load("pt_core_news_lg")
+        if lang == "pt":
+            if not(spacy.util.is_package('pt_core_news_lg')):
+                spacy.cli.download('pt_core_news_lg')
 
 
-        self.nlp.tokenizer = custom_tokenizer(self.nlp)
+            self.nlp = spacy.load("pt_core_news_lg")
+
+
+            self.nlp.tokenizer = custom_tokenizer(self.nlp)
+        else:
+            if not (spacy.util.is_package('en_core_web_lg')):
+                spacy.cli.download('en_core_web_lg')
+
+            self.nlp = spacy.load('en_core_web_lg')
 
         self.file_lst = []
         self.ann_ref = {}
@@ -136,12 +145,13 @@ class ReadBrat(read.Read):
         1, if el1 < el2, -1, if el1 > el2, and 0 if el1 is not adjacent to el2
         """
 
-        if "offset2_start" in el1 and "offset2_start" in el2:
+        if "offset2" in el1 and "offset2" in el2:
             return 0
 
         (left1_start, left1_end) = self._get_left_span(el1)
         (left2_start, left2_end) = self._get_left_span(el2)
 
+
         (right1_start, right1_end) = self._get_right_span(el1)
         (right2_start, right2_end) = self._get_right_span(el2)
 
@@ -162,7 +172,7 @@ class ReadBrat(read.Read):
     def _build_merged_element(self, el1, el2):
 
         new_el = {}
-        if "offset2_start" in el1:
+        if "offset2" in el1:
 
             new_el["offset1"] = (el1["offset1"][0], el1["offset1"][1])
 
@@ -173,7 +183,7 @@ class ReadBrat(read.Read):
             # the union if make in the first offset
             new_el["offset1"] = (el1["offset1"][0], el2["offset1"][1])
             if "offset2" in el2:
-                new_el["offset2"] = (el2["offset2"][0], el2["offset2_end"][1])
+                new_el["offset2"] = (el2["offset2"][0], el2["offset2"][1])
 
         new_el["value"] = el1["value"] + " " + el2["value"]
         new_el["id"] = el1["id"]
@@ -237,13 +247,38 @@ class ReadBrat(read.Read):
             if ent_type in LINK_TYPES:
                 new_ann[ent_type] = ann[ent_type]
                 continue # if it is a link, just ignore it
-
+            #if "offset2_end" not in ann[ent_type]:
+            #    print("-->", ann[ent_type])
             new_ann_ent = self.merge_span(ann[ent_type])
             new_ann[ent_type] = new_ann_ent
 
         return new_ann
 
-    def read_annotation_file(self, file_ann, merge_entities=True):
+    def get_attribute_value(self, element_id, attr_name):
+        if attr_name in self.ann_ref:
+            if element_id in self.ann_ref[attr_name]:
+                return self.ann_ref[attr_name][element_id]
+
+    def get_relation(self, element_a, element_b):
+        """
+        Get the relation name between two entities (their ids)
+        @param element_a: String with the id of an entity
+        @param element_b: String with the id of an entity
+        @return: String with relation name, NOne if there is no relation between
+        elements A and B
+        """
+
+        annotation_types = list(self.ann_ref.keys())
+        annotation_links_types = [ann_type for ann_type in annotation_types if "LINK" in ann_type]
+
+        for relation_name in  annotation_links_types:
+            if element_a in self.ann_ref[relation_name]:
+                if element_b in self.ann_ref[relation_name][element_a]:
+                    return relation_name
+
+
+
+    def read_annotation_file(self, file_ann, merge_entities=False):
         """
         It reads only the annotation file, then returns
         the processed tokens as TokenCorpus type
@@ -254,15 +289,26 @@ class ReadBrat(read.Read):
         """
 
         ann = {"Event": [], "Actor": [], "Time": [], "TIME_X3": [], \
-                "ACTOR": [], "Participant": [],"SRLINK":[]}
+                "ACTOR": [], "Participant": [],"SRLINK":[],"OLINK":[],"TLINK":[]}
+        ann_ref = {}  # annotation by reference
+        # the pattern of a reference for an argument of a relation
+        ARG_REL = re.compile("T\d+")
 
         with open(file_ann, "r") as fd:
+            last_line_type = None
             for line in fd:
                 ann_type = None
+                line_toks = line.split()
                 if line[0] != '#':
 
-                    line_toks = line.split()
+                    if len(line_toks) < 3:
+                        if line[0] != 'T' and line[0] != 'A' and line[0] != 'R' and last_line_type in ann.keys():
+                            ann[last_line_type][-1]["value"] += " " + line.strip()
+                        else:
+                            raise Exception(f"Invalid standoff line:{line}")
+                        continue
                     ann_type = line_toks[1]
+                    last_line_type = ann_type
 
                     if line[0] == 'T':
                         # TODO: catalogar aqui todos as entidades em um mapeamento
@@ -271,20 +317,29 @@ class ReadBrat(read.Read):
                             ann[ann_type] = []
 
                         if ';' in line_toks[3]:
-                            # this situation is when the annotation of th event
-                            # is two segments not adjacents
-                            offset1_start = int(line_toks[2])
-                            offset1_end = int(line_toks[4])
 
-                            offset2 = line_toks[3].split(';')
-                            offset2_start = int(offset2[0])
-                            offset2_end = int(offset2[1])
+                            start_range = int(line_toks[2])
+                            offset_lst = []
 
-                            value = " ".join(line_toks[5:])
+                            idx_toks = 3
+                            while ';' in line_toks[idx_toks]:
+                                tmp = line_toks[idx_toks].split(';')
+                                end_range = int(tmp[0])
+                                offset_lst.append((start_range, end_range))
 
-                            ann[ann_type].append({"id":line_toks[0],"offset1": (offset1_start, offset1_end), \
-                                              "offset2": (offset2_start, offset2_end), \
-                                              "value": value})
+                                start_range = int(tmp[1])
+                                idx_toks += 1
+
+                            end_range = int(line_toks[idx_toks])
+                            offset_lst.append((start_range, end_range))
+
+                            segment_ann = {"id":line_toks[0]} # annotation multi-segment annotation
+                            for idx_offset, (start_range, end_range) in enumerate(offset_lst):
+                                segment_ann[f"offset{idx_offset}"] = (start_range, end_range)
+
+                            value = " ".join(line_toks[idx_toks + 1:])
+                            segment_ann["value"] = value
+                            ann[ann_type].append(segment_ann)
 
                         else:
                             offset1_start = int(line_toks[2])
@@ -295,50 +350,87 @@ class ReadBrat(read.Read):
                             ann[ann_type].append({"id":line_toks[0],"offset1": (offset1_start, offset1_end), \
                                               "value": value})
                     elif line[0] == 'R':
-                        if ann_type.startswith("SEMROLE") or ann_type.startswith("SRLINK"):
-                            e1 = line_toks[2].split(":")[1] # entity 1
-                            e2 = line_toks[3].split(":")[1] # entity 2
-                            ann["SRLINK"].append({"id":line_toks[0],"args":(e1, e2)})
+                        e1 = line_toks[2].split(":")[1]  # entity 1
+                        e2 = line_toks[3].split(":")[1]  # entity 2
+                        rel_id = line_toks[0]
+                        if ann_type in ann:
+                            ann[ann_type].append({"id":rel_id,"args":(e1, e2)})
+                        else:
+                            ann[ann_type] = [{"id": rel_id, "args": (e1, e2)}]
 
-                    elif line[0] == 'A': # for now ignore attributes
-
-                        ref = line[2]
-                        attr_name = line[1]
-                        attr_value = line[3]
 
                         if ann_type in self.ann_ref:
-                            # ann_type keeps the last type of annotation
-                            if ref in self.ann_ref[ann_type]:
-                                self.ann_ref[ann_type][ref][attr_name] =  attr_value
+                            if e1 in self.ann_ref[ann_type]:
+                                self.ann_ref[ann_type][e1].append(e2)
                             else:
-                                self.ann_ref[ann_type][ref] = {attr_name: attr_value}
+                                self.ann_ref[ann_type][e1] = [e2]
                         else:
-                            self.ann_ref[ann_type] = {ref:{attr_name:attr_value}}
+                            self.ann_ref[ann_type] = {e1: [e2]}
+
+                    elif line[0] == 'A':
+
+                        ref = line_toks[2]
+                        if ref in ann_ref:
+                            # ann_ref[ref][1].append((line[1], line[3]))
+                            if len(line_toks) > 3:
+                                ann_ref[ref][line_toks[1]] = line_toks[3]
+                            else:
+                                ann_ref[ref][line_toks[1]] = line_toks[1]
+
+                        else:
+                            ann_ref[ref] = {line_toks[1]: line_toks[3]}
+                    elif line[0] == '#':
+                            # it is a note about the annotation,
+                            # insert it as an attribute
+                            ann_type = line_toks[1]
+                            ref = line_toks[2]
+                            ann_info = " ".join(line_toks[3:])
+
+                            if ref in ann_ref:
+                                if ann_type in ann_ref[ref]:
+                                    ann_ref[ref][ann_type].append(ann_info)
+                                else:
+                                    ann_ref[ref][ann_type] = [ann_info]
+                            else:
+                                ann_ref[ref] = {ann_type: [ann_info]}
+
         if merge_entities:
             return self.merge_entity_span(ann)
         else:
             return ann
 
-    def process(self, data_dir):
+    def process(self, data_dir, split=None):
         """
         It reads a set of files of annotations and text, then returns
         the processed tokens as TokenCorpus type
 
         @param string: path of data to gather and process
+        @param string: a file name that contains a list of the files in data dir that
+        should be processed by this reader
 
         @return [[TokenCorpus]]: a list of lists of tokens
         """
         # process the data corpus
         # and return a list of tokens
+        split_lst = []
+        if split is not None:
+            with open(split, "r") as fd:
+                split_lst = fd.readlines()
+                split_lst = [line.replace("\n","") for line in split_lst]
 
         data_tokens = []
 
         for dirpath, dirnames, filenames in os.walk(data_dir):
             for f in filenames:
+
                 if f.endswith(".ann"):
 
                     p = Path(f)
+                    if split_lst != [] and not(p.stem + ".txt" in split_lst):
+                        continue
+                 
                     fullname = os.path.join(data_dir, p.stem)
+                    #print("-->", fullname)
                     token_lst = self.process_file(fullname)
                     self.file_lst.append(fullname + ".txt")
 
@@ -370,11 +462,15 @@ class ReadBrat(read.Read):
 
     
         sent_id = -1
+        clause_id = 0
         for tok_idx, tok in enumerate(doc):
-
             if tok.is_sent_start:
                 sent_id += 1
 
+            # Check if the token marks the end of a clause
+            if tok.dep_ in ["punct", "mark"]:
+                clause_id += 1
+
             
             mytok = token_corpus.TokenCorpus()
             mytok.id = count
@@ -387,18 +483,20 @@ class ReadBrat(read.Read):
             mytok.head_lemma = tok.head.lemma_
             mytok.offset = tok.idx
             mytok.sent_id = sent_id
+            mytok.clause_id = clause_id
+            #print("-->",clause_id, sent_id)
 
 
-
-            ans = self.search_all_idx(tok.idx, ann_idx)
+            ans = self.search_all_idx(tok.idx, tok.idx + len(tok.text), ann_idx)
 
             # TODO: it is necessary perform a more efficient search
             # a possible subtoken annotation
             # perform more than one search, and build a list of id's
-            ans_sub = self.search_subtoken(tok.idx, tok.idx + len(tok.text), \
+            if len(tok.text.strip()) != 0 and tok.text not in ".!?-,":
+                ans_sub = self.search_subtoken(tok.idx, tok.idx + len(tok.text) - 1, \
                         ann_idx)
 
-            ans = ans.union(ans_sub)
+                ans = ans.union(ans_sub)
 
             if len(ans) != 0:
                 # annotations in token
@@ -414,7 +512,8 @@ class ReadBrat(read.Read):
 
                     for ref, ann_type, elems in ann[(t0, t1)]:
 
-                        mytok.attr.append((ann_type, ann_ref[ref]))
+                        mytok.attr.append((ann_type,ann_ref[ref]))
+                        mytok.ann_offset.append((t0,t1))
                         mytok.id_ann.append(ref) 
                         # all the annotations ids that a token can be assigned
                         ref_lst_tok.add(ref)
@@ -460,7 +559,7 @@ class ReadBrat(read.Read):
         ARG_REL = re.compile("T\d+") 
 
         with open(file_ann, "r") as fd:
-            
+
             for line in fd:
                 line_toks = line.split()
                 if line[0] != '#' and line[0] == 'T':
@@ -469,11 +568,33 @@ class ReadBrat(read.Read):
                     if ';' in line_toks[3]:
                         # this situation is when the annotation of th event
                         # is two segments not adjacents
-                        tmp = line_toks[3].split(';')
-                        ann[(int(line_toks[2]), int(tmp[0]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
-                        ann[(int(tmp[1]), int(line_toks[4]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
-                        ann_ref[line[0]] = (
-                            [(int(line_toks[2]), int(tmp[0])), (int(tmp[1]), int(line_toks[4]))], [])
+                        start_range = int(line_toks[2])
+                        offset_lst = []
+
+                        idx_toks = 3
+                        while ';' in line_toks[idx_toks]:
+
+                            tmp = line_toks[idx_toks].split(';')
+                            end_range = int(tmp[0])
+                            offset_lst.append((start_range, end_range))
+
+                            start_range = int(tmp[1])
+                            idx_toks += 1
+
+                        end_range = int(line_toks[idx_toks])
+                        offset_lst.append((start_range, end_range))
+
+                        for start_range, end_range in offset_lst:
+                            if (start_range, end_range) in ann:
+                                ann[(start_range, end_range)].append((line_toks[0], line_toks[1], line_toks[idx_toks + 1:]))
+                            else:
+                                ann[(start_range, end_range)] = [(line_toks[0], line_toks[1], line_toks[idx_toks + 1:])]
+
+                        #ann[(int(line_toks[2]), int(tmp[0]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
+                        #ann[(int(tmp[1]), int(line_toks[4]))] = [(line_toks[0], line_toks[1], line_toks[5:])]
+                        #ann_ref[line[0]] = (
+                        #    [(int(line_toks[2]), int(tmp[0])), (int(tmp[1]), int(line_toks[4]))], [])
+                        ann_ref[line[0]] = (offset_lst, [])
 
 
                         ann_ref[line_toks[0]] = {}
@@ -519,9 +640,27 @@ class ReadBrat(read.Read):
                     ref = line_toks[2]
                     if ref in ann_ref:
                         #ann_ref[ref][1].append((line[1], line[3]))
-                        ann_ref[ref][line_toks[1]] = line_toks[3]
+                        if len(line_toks) > 3:
+                            ann_ref[ref][line_toks[1]] = line_toks[3]
+                        else:
+                            ann_ref[ref][line_toks[1]] = line_toks[1]
+
                     else:
                         ann_ref[ref] = {line_toks[1]:line_toks[3]}
+                elif line[0] == '#':
+                    # it is a note about the annotation,
+                    # insert it as an attribute
+                    ann_type = line_toks[1]
+                    ref = line_toks[2]
+                    ann_info = " ".join(line_toks[3:])
+
+                    if ref in ann_ref:
+                        if ann_type in ann_ref[ref]:
+                            ann_ref[ref][ann_type].append(ann_info)
+                        else:
+                            ann_ref[ref][ann_type] = [ann_info]
+                    else:
+                        ann_ref[ref] = {ann_type: [ann_info]}
 
         idx_lst = ann.keys()
 
@@ -534,10 +673,9 @@ class ReadBrat(read.Read):
 
         with open(file_txt, "r") as fd:
 
-            doc = self.nlp(fd.read())
+            doc = self.nlp(fd.read().strip())
             token_lst, ref2tok = self.extract_token_corpus(doc, idx_lst, ann, ann_ref)
 
-
         # update relation field for each token
         # each relation field is a list of TokenRelation object, which 
         # points to a TokenCorpus and specifies the type of relation
@@ -551,13 +689,15 @@ class ReadBrat(read.Read):
 
                         # the relation for the reference ref
                         for rel_id, rel_type, ref_arg,argn in ann_rel[ref]:
-                            rel_obj = token_corpus.TokenRelation(rel_id, ref2tok[ref_arg], rel_type, argn, ref_arg)
-                            mytok.relations.append(rel_obj)
-
+                            try:
+                                rel_obj = token_corpus.TokenRelation(rel_id, ref2tok[ref_arg], rel_type, argn, ref_arg)
+                                mytok.relations.append(rel_obj)
+                            except KeyError:
+                                print(f"Warning: The {rel_id} was not included in the doc since {ref_arg} was not found as annotation.")
 
         return token_lst
 
-    def search_all_idx(self, idx, idx_lst):
+    def search_all_idx(self, idx, idx_end, idx_lst):
         """
         Since a token can be annotated with multiple id's, it is necessary 
         to perform multiple search for each one of these id's.
@@ -591,7 +731,8 @@ class ReadBrat(read.Read):
 
         for pos, (start, end) in enumerate(idx_lst):
             
-            if start >= tok_start and end <= tok_end:
+            #if start >= tok_start and end <= tok_end:
+            if not (tok_end < start or tok_start > end):
                 ans.add(pos)
 
         return ans
@@ -634,6 +775,7 @@ class ReadBrat(read.Read):
     def __process_annotations(self, file_ann, ann, ann_ref):
         pass
 
+
     def fileToColumnFormat(self, ann_file, output_file):
         """
         Convert only one file to the column format.
